REGIONS = {'Norte':        ['Amazonas', 'Roraima', 'Amapá', 'Pará', 'Tocantins', 'Rondônia', 'Acre'],
           'Nordeste':     ['Maranhão', 'Piauí', 'Ceará', 'Rio Grande do Norte', 'Pernambuco', 'Paraíba', 'Sergipe',
                            'Alagoas', 'Bahia'],
           'Centro-Oeste': ['Mato Grosso', 'Mato Grosso do Sul', 'Goiás', 'Distrito Federal'],
           'Sudeste':      ['São Paulo', 'Rio de Janeiro', 'Espírito Santo', 'Minas Gerais'],
           'Sul':          ['Paraná', 'Rio Grande do Sul', 'Santa Catarina']
           }
STATES = {'Rondônia':       'RO', 'Acre': 'AC', 'Amazonas': 'AM', 'Roraima': 'RR', 'Pará': 'PA', 'Amapá': 'AP',
          'Tocantins':      'TO', 'Maranhão': 'MA', 'Piauí': 'PI', 'Ceará': 'CE', 'Rio Grande do Norte': 'RN',
          'Paraíba':        'PB', 'Pernambuco': 'PE', 'Alagoas': 'AL', 'Sergipe': 'SE', 'Bahia': 'BA',
          'Minas Gerais':   'MG', 'Espírito Santo': 'ES', 'Rio de Janeiro': 'RJ', 'São Paulo': 'SP', 'Paraná': 'PR',
          'Santa Catarina': 'SC', 'Rio Grande do Sul': 'RS', 'Mato Grosso do Sul': 'MS', 'Mato Grosso': 'MT',
          'Goiás':          'GO', 'Distrito Federal': 'DF'}